/** **********************************************************************
 * Specialized two-input nodes for negated patterns
 *
 * NOTE: CLIPS behaves in a surprising way which I'm following here.
 * Given this defrule:
 * <PRE>
 * (defrule test-1
 *  (bar)       
 *  (not (foo))
 *  =>
 *  (printout t "not foo"))
 *  
 * CLIPS behaves this way:
 * 
 * (watch activations)
 * (assert (bar))
 * ==> Activation 0 test-1
 * (assert (foo))
 * <== Activation 0 test-1
 * (retract (foo))
 * ==> Activation 0 test-1
 * 
 * This is not surprising yet. Here's the funky part
 * 
 * (run)
 * "not foo"
 * (assert (foo))
 * (retract (foo))
 * ==> Activation 0 test-1
 *
 * The rule fires,  and all that's required to fire it again is for the
 * "not-ness" to be removed and replaced; the (bar) fact does not need to
 * be replaced. This obviously falls out of the implementation; it makes things
 * easy!
 *
 * </PRE>
*
 * $Id: NodeNot2.java,v 1.9 2001/09/24 19:14:40 ejfried Exp $
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */

package jess;
import java.io.*;
import java.util.ArrayList;

/**
 * (C) 1997 Ernest J. Friedman-Hill and Sandia National Laboratories
 * @author Ernest J. Friedman-Hill
 */
class NodeNot2 extends Node2 implements Serializable
{
  /**
   * @param engine 
   */
  NodeNot2(Rete engine, int hashkey) throws JessException
  {
    super(engine, false, hashkey);
  }

  boolean callNodeLeft(Token token) throws JessException
  {  
    if (token.m_tag == RU.ADD || token.m_tag == RU.UPDATE)
      token = Rete.getFactory().newToken(token);

    return super.callNodeLeft(token);
  }
    
  /**
   * Run all the tests on a given (left) token and every token in the
   * right memory. Every time a right token *passes* the tests, increment 
   * the left token's negation count; at the end, if the
   * left token has a zero count, pass it through.
   * 
   * The 'nullToken' contains a fact used as a placeholder for the 'not' CE.
   * @param lt 
   * @param th 
   * @exception JessException 
   * @return 
   */
  boolean runTestsVaryRight(Token lt, TokenTree th) throws JessException 
  {
    if (lt.m_tag != RU.REMOVE)      
      super.runTestsVaryRight(lt, th);
    
    if (lt.m_negcnt == 0) {
        Token newToken = Rete.getFactory().newToken(lt, Fact.getNullFact());
        newToken.updateTime(Rete.getEngine());
        m_matches++;
        passAlong(newToken);
        if (m_logicalDepends != null)
            removeLogicalSupportFrom(newToken);
      }
    return false;
  }

  /**
   * @param lt 
   * @param th 
   * @exception JessException 
   */
  boolean doRunTestsVaryRight(Token lt, TokenTree th)
       throws JessException
  {
    if (th == null)
      return false;
    
    Context context = new Context(Rete.getEngine().getGlobalContext());
    for (int j=0; j<th.m_hash; j++)
      doRunTestsVaryRight(lt, th.m_tokens[j], context);

    return false;
  }
  
  boolean doRunTestsVaryRight(Token lt, TokenVector v, Context context) throws JessException
  {
    if (v != null)
    {
      int size = v.size();
      if (size > 0)
        {
          int ntests = m_nTests;
          context.setToken(lt);
          
          for (int i=0; i<size; i++)
            {
              context.setFact(v.elementAt(i).topFact());
              
              if (ntests == 0 || runTests(ntests, context))
                lt.m_negcnt++;
            }
        }
    }
    return false;

  }

  /**
   * Run all the tests on a given (right) token and every token in the
   * left memory. For the true ones, increment (or decrement) the appropriate
   * negation counts. Any left token which transitions to zero gets passed
   * along.
   * @param rt 
   * @param th 
   * @exception JessException 
   * @return 
   */

  boolean doRunTestsVaryLeft(Token rt, TokenTree th) throws JessException 
  {
    if (th == null)
      return false;

    Context context = new Context(Rete.getEngine().getGlobalContext()); 
    for (int j=0; j<th.m_hash; j++)
      {
        TokenVector v = th.m_tokens[j];
        doRunTestsVaryLeft(rt, v, context);
      }

    return false;
  }

  boolean doRunTestsVaryLeft(Token rt, TokenVector v, Context context) throws JessException
  {
    if (v != null)
      {
        int size = v.size();
        if (size > 0)
          {
            int ntests = m_nTests;
            int tag = rt.m_tag;
            context.setFact(rt.topFact());
            for (int i=0; i<size; i++)
              {
                Token lt = v.elementAt(i);
                context.setToken(lt);
                
                if (ntests == 0 || runTests(ntests, context)) {
                  if (tag == RU.ADD || tag == RU.UPDATE) 
                    {
                      // retract any activation due to the left token
                      Token nt2 = Rete.getFactory().newToken(lt, Fact.getNullFact());
                      nt2.updateTime(Rete.getEngine());
                      nt2.m_tag = RU.REMOVE;
                      passAlong(nt2);
                      if (m_logicalDepends != null)
                          removeLogicalSupportFrom(nt2);
                      lt.m_negcnt++;
                    } 
                  else if (--lt.m_negcnt == 0) 
                    {
                      // pass along the revitalized left token
                      Token nt2 = Rete.getFactory().newToken(lt, Fact.getNullFact());
                      nt2.updateTime(Rete.getEngine());
                      passAlong(nt2);
                      if (m_logicalDepends != null)
                          removeLogicalSupportFrom(nt2);
                    }
                  else if (lt.m_negcnt < 0)
                    throw new JessException("NodeNot2.RunTestsVaryLeft",
                                            "Corrupted Negcnt (< 0)",
                                            "");
                }
              }
          }
      }
    return false;
  }
  /**
   * Describe myself
   * @return 
   */
  public String toString() 
  {
    StringBuffer sb = new StringBuffer(256);
    sb.append("[NodeNot2 ntests=");
    sb.append(m_nTests);
    sb.append(" ");
    for (int i=0; i<m_nTests; i++)
      {
        sb.append(m_tests[i].toString());
        sb.append(" ");
      }
    sb.append(";usecount = ");
    sb.append(m_usecount);
    sb.append("]");
    return sb.toString();
  }
}
